import copy
import typing
import logging

import hyperopt

import matchzoo as mz
from matchzoo.engine.base_metric import BaseMetric
from .callbacks.callback import Callback


class Tuner(object):
    """
   Model hyper-parameters tuner.

    `model.params.hyper_space` reprensents the model's hyper-parameters
    search space, which is the cross-product of individual hyper parameter's
    hyper space. When a `Tuner` builds a model, for each hyper parameter in
    `model.params`, if the hyper-parameter has a hyper-space, then a sample
    will be taken in the space. However, if the hyper-parameter does not
    have a hyper-space, then the default value of the hyper-parameter will
    be used.

    See `tutorials/model_tuning.ipynb` for a detailed walkthrough on usage.

    :param params: A completed parameter table to tune. Usually `model.params`
        of the desired model to tune. `params.completed()` should be `True`.
    :param train_data: Training data to use. Either a preprocessed `DataPack`,
        or a `DataGenerator`.
    :param test_data: Testing data to use. A preprocessed `DataPack`.
    :param fit_kwargs: Extra keyword arguments to pass to `fit`.
        (default: `dict(epochs=10, verbose=0)`)
    :param evaluate_kwargs: Extra keyword arguments to pass to `evaluate`.
    :param metric: Metric to tune upon. Must be one of the metrics in
        `model.params['task'].metrics`. (default: the first metric in
        `params.['task'].metrics`.
    :param mode: Either `maximize` the metric or `minimize` the metric.
        (default: 'maximize')
    :param num_runs: Number of runs. Each run takes a sample in
        `params.hyper_space` and build a model based on the sample.
        (default: 10)
    :param callbacks: A list of callbacks to handle. Handled sequentially
        at every callback point.
    :param verbose: Verbosity. (default: 1)

    Example:
        >>> import matchzoo as mz
        >>> train = mz.datasets.toy.load_data('train')
        >>> dev = mz.datasets.toy.load_data('dev')
        >>> prpr = mz.models.DenseBaseline.get_default_preprocessor()
        >>> train = prpr.fit_transform(train, verbose=0)
        >>> dev = prpr.transform(dev, verbose=0)
        >>> model = mz.models.DenseBaseline()
        >>> model.params['input_shapes'] = prpr.context['input_shapes']
        >>> model.params['task'] = mz.tasks.Ranking()
        >>> tuner = mz.auto.Tuner(
        ...     params=model.params,
        ...     train_data=train,
        ...     test_data=dev,
        ...     num_runs=1,
        ...     verbose=0
        ... )
        >>> results = tuner.tune()
        >>> sorted(results['best'].keys())
        ['#', 'params', 'sample', 'score']

    """

    def __init__(
        self,
        params: 'mz.ParamTable',
        train_data: typing.Union[mz.DataPack, mz.DataGenerator],
        test_data: typing.Union[mz.DataPack, mz.DataGenerator],
        fit_kwargs: dict = None,
        evaluate_kwargs: dict = None,
        metric: typing.Union[str, BaseMetric] = None,
        mode: str = 'maximize',
        num_runs: int = 10,
        callbacks: typing.List[Callback] = None,
        verbose=1
    ):
        """Tuner."""
        if fit_kwargs is None:
            fit_kwargs = dict(epochs=10, verbose=0)
        if evaluate_kwargs is None:
            evaluate_kwargs = {}
        if callbacks is None:
            callbacks = []

        self._validate_params(params)
        metric = metric or params['task'].metrics[0]
        self._validate_data(train_data)
        self._validate_data(test_data)
        self._validate_kwargs(fit_kwargs)
        self._validate_kwargs(evaluate_kwargs)
        self._validate_mode(mode)
        self._validate_metric(params, metric)
        self._validate_callbacks(callbacks)

        self.__curr_run_num = 0

        # these variables should not change within the same `tune` call
        self._params = params
        self._train_data = train_data
        self._test_data = test_data
        self._fit_kwargs = fit_kwargs
        self._evaluate_kwargs = evaluate_kwargs
        self._metric = metric
        self._mode = mode
        self._num_runs = num_runs
        self._callbacks = callbacks
        self._verbose = verbose

    def tune(self):
        """
        Start tuning.

        Notice that `tune` does not affect the tuner's inner state, so each
        new call to `tune` starts fresh. In other words, hyperspaces are
        suggestive only within the same `tune` call.
        """
        if self.__curr_run_num != 0:
            print(
                """WARNING: `tune` does not affect the tuner's inner state, so
                each new call to `tune` starts fresh. In other words,
                hyperspaces are suggestive only within the same `tune` call."""
            )
        self.__curr_run_num = 0
        logging.getLogger('hyperopt').setLevel(logging.CRITICAL)

        trials = hyperopt.Trials()

        self._fmin(trials)

        return {
            'best': trials.best_trial['result']['mz_result'],
            'trials': [trial['result']['mz_result'] for trial in trials.trials]
        }

    def _fmin(self, trials):
        # new version of hyperopt has keyword argument `show_progressbar` that
        # breaks doctests, so here's a workaround
        fmin_kwargs = dict(
            fn=self._run,
            space=self._params.hyper_space,
            algo=hyperopt.tpe.suggest,
            max_evals=self._num_runs,
            trials=trials
        )
        try:
            hyperopt.fmin(
                **fmin_kwargs,
                show_progressbar=False
            )
        except TypeError:
            hyperopt.fmin(**fmin_kwargs)

    def _run(self, sample):
        self.__curr_run_num += 1

        # build start
        self._handle_callbacks_run_start(sample)

        # build model
        params = self._create_full_params(sample)
        model = params['model_class'](params=params)
        model.build()
        model.compile()
        self._handle_callbacks_build_end(model)

        # fit & evaluate
        self._fit_model(model)
        lookup = self._evaluate_model(model)
        score = lookup[self._metric]

        # collect result
        # this result is for users, visible outside
        mz_result = {
            '#': self.__curr_run_num,
            'params': params,
            'sample': sample,
            'score': score
        }

        self._handle_callbacks_run_end(model, mz_result)

        if self._verbose:
            self._log_result(mz_result)

        return {
            # these two items are for hyperopt
            'loss': self._fix_loss_sign(score),
            'status': hyperopt.STATUS_OK,

            # this item is for storing matchzoo information
            'mz_result': mz_result
        }

    def _create_full_params(self, sample):
        params = copy.deepcopy(self._params)
        params.update(sample)
        return params

    def _handle_callbacks_run_start(self, sample):
        for callback in self._callbacks:
            callback.on_run_start(self, sample)

    def _handle_callbacks_build_end(self, model):
        for callback in self._callbacks:
            callback.on_build_end(self, model)

    def _handle_callbacks_run_end(self, model, result):
        for callback in self._callbacks:
            callback.on_run_end(self, model, result)

    def _fit_model(self, model):
        if isinstance(self._train_data, mz.DataPack):
            x, y = self._train_data.unpack()
            model.fit(x, y, **self._fit_kwargs)
        elif isinstance(self._train_data, mz.DataGenerator):
            model.fit_generator(self._train_data, **self._fit_kwargs)
        else:
            raise ValueError(f"Invalid data type: `train_data`."
                             f"{type(self._train_data)} received."
                             f"Must be one of `DataPack` and `DataGenerator`.")

    def _evaluate_model(self, model):
        if isinstance(self._test_data, mz.DataPack):
            x, y = self._test_data.unpack()
            return model.evaluate(x, y, **self._evaluate_kwargs)
        elif isinstance(self._test_data, mz.DataGenerator):
            return model.evaluate_generator(self._test_data,
                                            **self._evaluate_kwargs)
        else:
            raise ValueError(f"Invalid data type: `test_data`."
                             f"{type(self._test_data)} received."
                             f"Must be one of `DataPack` and `DataGenerator`.")

    def _fix_loss_sign(self, loss):
        if self._mode == 'maximize':
            loss = -loss
        return loss

    @classmethod
    def _log_result(cls, result):
        print(f"Run #{result['#']}")
        print(f"Score: {result['score']}")
        print(result['params'])
        print()

    @property
    def params(self):
        """`params` getter."""
        return self._params

    @params.setter
    def params(self, value):
        """`params` setter."""
        self._validate_params(value)
        self._validate_metric(value, self._metric)
        self._params = value

    @property
    def train_data(self):
        """`train_data` getter."""
        return self._train_data

    @train_data.setter
    def train_data(self, value):
        """`train_data` setter."""
        self._validate_data(value)
        self._train_data = value

    @property
    def test_data(self):
        """`test_data` getter."""
        return self._test_data

    @test_data.setter
    def test_data(self, value):
        """`test_data` setter."""
        self._validate_data(value)
        self._test_data = value

    @property
    def fit_kwargs(self):
        """`fit_kwargs` getter."""
        return self._fit_kwargs

    @fit_kwargs.setter
    def fit_kwargs(self, value):
        """`fit_kwargs` setter."""
        self._validate_kwargs(value)
        self._fit_kwargs = value

    @property
    def evaluate_kwargs(self):
        """`evaluate_kwargs` getter."""
        return self._evaluate_kwargs

    @evaluate_kwargs.setter
    def evaluate_kwargs(self, value):
        """`evaluate_kwargs` setter."""
        self._validate_kwargs(value)
        self._evaluate_kwargs = value

    @property
    def metric(self):
        """`metric` getter."""
        return self._metric

    @metric.setter
    def metric(self, value):
        """`metric` setter."""
        self._validate_metric(self._params, value)
        self._metric = value

    @property
    def mode(self):
        """`mode` getter."""
        return self._mode

    @mode.setter
    def mode(self, value):
        """`mode` setter."""
        self._validate_mode(value)
        self._mode = value

    @property
    def num_runs(self):
        """`num_runs` getter."""
        return self._num_runs

    @num_runs.setter
    def num_runs(self, value):
        """`num_runs` setter."""
        self._validate_num_runs(value)
        self._num_runs = value

    @property
    def callbacks(self):
        """`callbacks` getter."""
        return self._callbacks

    @callbacks.setter
    def callbacks(self, value):
        """`callbacks` setter."""
        self._validate_callbacks(value)
        self._callbacks = value

    @property
    def verbose(self):
        """`verbose` getter."""
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        """`verbose` setter."""
        self._verbose = value

    @classmethod
    def _validate_params(cls, params):
        if not isinstance(params, mz.ParamTable):
            raise TypeError("Only accepts a `ParamTable` instance.")
        if not params.hyper_space:
            raise ValueError("Parameter hyper-space empty.")
        if not params.completed():
            raise ValueError("Parameters not complete.")

    @classmethod
    def _validate_data(cls, train_data):
        if not isinstance(train_data, (mz.DataPack, mz.DataGenerator)):
            raise TypeError(
                "Only accepts a `DataPack` or `DataGenerator` instance.")

    @classmethod
    def _validate_kwargs(cls, kwargs):
        if not isinstance(kwargs, dict):
            raise TypeError('Only accepts a `dict` instance.')

    @classmethod
    def _validate_mode(cls, mode):
        if mode not in ('maximize', 'minimize'):
            raise ValueError('`mode` should be one of `maximize`, `minimize`.')

    @classmethod
    def _validate_metric(cls, params, metric):
        if metric not in params['task'].metrics:
            raise ValueError('Target metric does not exist in the task.')

    @classmethod
    def _validate_num_runs(cls, num_runs):
        if not isinstance(num_runs, int):
            raise TypeError('Only accepts an `int` value.')

    @classmethod
    def _validate_callbacks(cls, callbacks):
        for callback in callbacks:
            if not isinstance(callback, Callback):
                if issubclass(callback, Callback):
                    raise TypeError("Make sure to instantiate the callbacks.")
                raise TypeError('Only accepts a `callbacks` instance.')



"""ConvKNRM model."""

import keras
import keras.backend as K

from .knrm import KNRM
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class ConvKNRM(KNRM):
    """
    ConvKNRM model.

    Examples:
        >>> model = ConvKNRM()
        >>> model.params['embedding_input_dim'] = 10000
        >>> model.params['embedding_output_dim'] = 300
        >>> model.params['embedding_trainable'] = True
        >>> model.params['filters'] = 128
        >>> model.params['conv_activation_func'] = 'tanh'
        >>> model.params['max_ngram'] = 3
        >>> model.params['use_crossmatch'] = True
        >>> model.params['kernel_num'] = 11
        >>> model.params['sigma'] = 0.1
        >>> model.params['exact_sigma'] = 0.001
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    def get_default_params(cls):
        """Get default parameters."""
        params = super().get_default_params()
        params.add(Param(name='filters', value=128,
                         desc="The filter size in the convolution"
                              " layer."))
        params.add(Param(name='conv_activation_func', value='relu',
                         desc="The activation function in the "
                              "convolution layer."))
        params.add(Param(name='max_ngram', value=3,
                         desc="The maximum length of n-grams for the "
                              "convolution layer."))
        params.add(Param(name='use_crossmatch', value=True,
                         desc="Whether to match left n-grams and right "
                              "n-grams of different lengths"))
        return params

    def build(self):
        """Build model."""
        query, doc = self._make_inputs()

        embedding = self._make_embedding_layer()

        q_embed = embedding(query)
        d_embed = embedding(doc)

        q_convs = []
        d_convs = []
        for i in range(self._params['max_ngram']):
            c = keras.layers.Conv1D(
                self._params['filters'], i + 1,
                activation=self._params['conv_activation_func'],
                padding='same'
            )
            q_convs.append(c(q_embed))
            d_convs.append(c(d_embed))

        KM = []
        for qi in range(self._params['max_ngram']):
            for di in range(self._params['max_ngram']):
                # do not match n-gram with different length if use crossmatch
                if not self._params['use_crossmatch'] and qi != di:
                    continue
                q_ngram = q_convs[qi]
                d_ngram = d_convs[di]
                mm = keras.layers.Dot(axes=[2, 2],
                                      normalize=True)([q_ngram, d_ngram])

                for i in range(self._params['kernel_num']):
                    mu = 1. / (self._params['kernel_num'] - 1) + (2. * i) / (
                        self._params['kernel_num'] - 1) - 1.0
                    sigma = self._params['sigma']
                    if mu > 1.0:
                        sigma = self._params['exact_sigma']
                        mu = 1.0
                    mm_exp = self._kernel_layer(mu, sigma)(mm)
                    mm_doc_sum = keras.layers.Lambda(
                        lambda x: K.tf.reduce_sum(x, 2))(
                        mm_exp)
                    mm_log = keras.layers.Activation(K.tf.log1p)(mm_doc_sum)
                    mm_sum = keras.layers.Lambda(
                        lambda x: K.tf.reduce_sum(x, 1))(mm_log)
                    KM.append(mm_sum)

        phi = keras.layers.Lambda(lambda x: K.tf.stack(x, 1))(KM)
        out = self._make_output_layer()(phi)
        self._backend = keras.Model(inputs=[query, doc], outputs=[out])



"""Mean average precision metric for ranking."""
import numpy as np

from matchzoo.engine.base_metric import BaseMetric, sort_and_couple


class MeanAveragePrecision(BaseMetric):
    """Mean average precision metric."""

    ALIAS = ['mean_average_precision', 'map']

    def __init__(self, threshold: float = 0.):
        """
        :class:`MeanAveragePrecision` constructor.

        :param threshold: The threshold of relevance degree.
        """
        self._threshold = threshold

    def __repr__(self):
        """:return: Formated string representation of the metric."""
        return f"{self.ALIAS[0]}({self._threshold})"

    def __call__(self, y_true: np.array, y_pred: np.array) -> float:
        """
        Calculate mean average precision.

        Example:
            >>> y_true = [0, 1, 0, 0]
            >>> y_pred = [0.1, 0.6, 0.2, 0.3]
            >>> MeanAveragePrecision()(y_true, y_pred)
            1.0

        :param y_true: The ground true label of each document.
        :param y_pred: The predicted scores of each document.
        :return: Mean average precision.
        """
        result = 0.
        pos = 0
        coupled_pair = sort_and_couple(y_true, y_pred)
        for idx, (label, score) in enumerate(coupled_pair):
            if label > self._threshold:
                pos += 1.
                result += pos / (idx + 1.)
        if pos == 0:
            return 0.
        else:
            return result / pos



import nltk

from .unit import Unit


class Tokenize(Unit):
    """Process unit for text tokenization."""

    def transform(self, input_: str) -> list:
        """
        Process input data from raw terms to list of tokens.

        :param input_: raw textual input.

        :return tokens: tokenized tokens as a list.
        """
        return nltk.word_tokenize(input_)



"""An implementation of DRMMTKS Model."""
import typing

import keras
import keras.backend as K

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class DRMMTKS(BaseModel):
    """
    DRMMTKS Model.

    Examples:
        >>> model = DRMMTKS()
        >>> model.params['embedding_input_dim'] = 10000
        >>> model.params['embedding_output_dim'] = 100
        >>> model.params['top_k'] = 20
        >>> model.params['mlp_num_layers'] = 1
        >>> model.params['mlp_num_units'] = 5
        >>> model.params['mlp_num_fan_out'] = 1
        >>> model.params['mlp_activation_func'] = 'tanh'
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(
            with_embedding=True,
            with_multi_layer_perceptron=True
        )
        params.add(Param(name='mask_value', value=-1,
                         desc="The value to be masked from inputs."))
        params['input_shapes'] = [(5,), (300,)]
        params.add(Param(
            'top_k', value=10,
            hyper_space=hyper_spaces.quniform(low=2, high=100),
            desc="Size of top-k pooling layer."
        ))
        return params

    def build(self):
        """Build model structure."""

        # Scalar dimensions referenced here:
        #   B = batch size (number of sequences)
        #   D = embedding size
        #   L = `input_left` sequence length
        #   R = `input_right` sequence length
        #   K = size of top-k

        # Left input and right input.
        # shape = [B, L]
        # shape = [B, R]
        query, doc = self._make_inputs()

        embedding = self._make_embedding_layer()
        # Process left input.
        # shape = [B, L, D]
        embed_query = embedding(query)
        # shape = [B, R, D]
        embed_doc = embedding(doc)
        # shape = [B, L]
        atten_mask = K.not_equal(query, self._params['mask_value'])
        # shape = [B, L]
        atten_mask = K.cast(atten_mask, K.floatx())
        # shape = [B, L, 1]
        atten_mask = K.expand_dims(atten_mask, axis=2)
        # shape = [B, L, 1]
        attention_probs = self.attention_layer(embed_query, atten_mask)

        # Matching histogram of top-k
        # shape = [B, L, R]
        matching_matrix = keras.layers.Dot(axes=[2, 2], normalize=True)(
            [embed_query,
             embed_doc])
        # shape = [B, L, K]
        effective_top_k = min(self._params['top_k'],
                              self.params['input_shapes'][0][0],
                              self.params['input_shapes'][1][0])
        matching_topk = keras.layers.Lambda(
            lambda x: K.tf.nn.top_k(x, k=effective_top_k, sorted=True)[0]
        )(matching_matrix)

        # Process right input.
        # shape = [B, L, 1]
        dense_output = self._make_multi_layer_perceptron_layer()(matching_topk)

        # shape = [B, 1, 1]
        dot_score = keras.layers.Dot(axes=[1, 1])(
            [attention_probs, dense_output])

        flatten_score = keras.layers.Flatten()(dot_score)

        x_out = self._make_output_layer()(flatten_score)
        self._backend = keras.Model(inputs=[query, doc], outputs=x_out)

    @classmethod
    def attention_layer(cls, attention_input: typing.Any,
                        attention_mask: typing.Any = None
                        ) -> keras.layers.Layer:
        """
        Performs attention on the input.

        :param attention_input: The input tensor for attention layer.
        :param attention_mask: A tensor to mask the invalid values.
        :return: The masked output tensor.
        """
        # shape = [B, L, 1]
        dense_input = keras.layers.Dense(1, use_bias=False)(attention_input)
        if attention_mask is not None:
            # Since attention_mask is 1.0 for positions we want to attend and
            # 0.0 for masked positions, this operation will create a tensor
            # which is 0.0 for positions we want to attend and -10000.0 for
            # masked positions.

            # shape = [B, L, 1]
            dense_input = keras.layers.Lambda(
                lambda x: x + (1.0 - attention_mask) * -10000.0,
                name="attention_mask"
            )(dense_input)
        # shape = [B, L, 1]
        attention_probs = keras.layers.Lambda(
            lambda x: keras.layers.activations.softmax(x, axis=1),
            output_shape=lambda s: (s[0], s[1], s[2]),
            name="attention_probs"
        )(dense_input)
        return attention_probs



from .unit import Unit
from .digit_removal import DigitRemoval
from .fixed_length import FixedLength
from .frequency_filter import FrequencyFilter
from .lemmatization import Lemmatization
from .lowercase import Lowercase
from .matching_histogram import MatchingHistogram
from .ngram_letter import NgramLetter
from .punc_removal import PuncRemoval
from .stateful_unit import StatefulUnit
from .stemming import Stemming
from .stop_removal import StopRemoval
from .tokenize import Tokenize
from .vocabulary import Vocabulary
from .word_hashing import WordHashing


def list_available() -> list:
    from matchzoo.utils import list_recursive_concrete_subclasses
    return list_recursive_concrete_subclasses(Unit)



from .naive import Naive
from .dssm import DSSM
from .cdssm import CDSSM
from .dense_baseline import DenseBaseline
from .arci import ArcI
from .arcii import ArcII
from .match_pyramid import MatchPyramid
from .knrm import KNRM
from .conv_knrm import ConvKNRM
from .duet import DUET
from .drmmtks import DRMMTKS
from .drmm import DRMM
from .anmm import ANMM
from .mvlstm import MVLSTM


def list_available() -> list:
    from matchzoo.engine.base_model import BaseModel
    from matchzoo.utils import list_recursive_concrete_subclasses
    return list_recursive_concrete_subclasses(BaseModel)



"""Classification task."""

from matchzoo.engine.base_task import BaseTask


class Classification(BaseTask):
    """Classification task.

    Examples:
        >>> classification_task = Classification(num_classes=2)
        >>> classification_task.metrics = ['precision']
        >>> classification_task.num_classes
        2
        >>> classification_task.output_shape
        (2,)
        >>> classification_task.output_dtype
        <class 'int'>
        >>> print(classification_task)
        Classification Task with 2 classes

    """

    def __init__(self, num_classes: int = 2, **kwargs):
        """Classification task."""
        super().__init__(**kwargs)
        if not isinstance(num_classes, int):
            raise TypeError("Number of classes must be an integer.")
        if num_classes < 2:
            raise ValueError("Number of classes can't be smaller than 2")
        self._num_classes = num_classes

    @property
    def num_classes(self) -> int:
        """:return: number of classes to classify."""
        return self._num_classes

    @classmethod
    def list_available_losses(cls) -> list:
        """:return: a list of available losses."""
        return ['categorical_crossentropy']

    @classmethod
    def list_available_metrics(cls) -> list:
        """:return: a list of available metrics."""
        return ['acc']

    @property
    def output_shape(self) -> tuple:
        """:return: output shape of a single sample of the task."""
        return self._num_classes,

    @property
    def output_dtype(self):
        """:return: target data type, expect `int` as output."""
        return int

    def __str__(self):
        """:return: Task name as string."""
        return f'Classification Task with {self._num_classes} classes'



import pytest

import matchzoo as mz


@pytest.mark.slow
def test_load_data():
    train_data = mz.datasets.wiki_qa.load_data('train', task='ranking')
    assert len(train_data) == 20360
    train_data, _ = mz.datasets.wiki_qa.load_data('train',
                                                  task='classification',
                                                  return_classes=True)
    assert len(train_data) == 20360

    dev_data = mz.datasets.wiki_qa.load_data('dev', task='ranking',
                                             filtered=False)
    assert len(dev_data) == 2733
    dev_data, tag = mz.datasets.wiki_qa.load_data('dev', task='classification',
                                                  filtered=True,
                                                  return_classes=True)
    assert len(dev_data) == 1126
    assert tag == [False, True]

    test_data = mz.datasets.wiki_qa.load_data('test', task='ranking',
                                              filtered=False)
    assert len(test_data) == 6165
    test_data, tag = mz.datasets.wiki_qa.load_data('test',
                                                   task='classification',
                                                   filtered=True,
                                                   return_classes=True)
    assert len(test_data) == 2341
    assert tag == [False, True]


@pytest.mark.slow
def test_load_snli():
    train_data, classes = mz.datasets.snli.load_data('train',
                                                     'classification',
                                                     return_classes=True)
    num_samples = 550146
    assert len(train_data) == num_samples
    x, y = train_data.unpack()
    assert len(x['text_left']) == num_samples
    assert len(x['text_right']) == num_samples
    assert y.shape == (num_samples, 4)
    assert classes == ['entailment', 'contradiction', 'neutral', '-']
    dev_data, classes = mz.datasets.snli.load_data('dev', 'classification',
                                                   return_classes=True)
    assert len(dev_data) == 10000
    assert classes == ['entailment', 'contradiction', 'neutral', '-']
    test_data, classes = mz.datasets.snli.load_data('test', 'classification',
                                                    return_classes=True)
    assert len(test_data) == 10000
    assert classes == ['entailment', 'contradiction', 'neutral', '-']

    train_data = mz.datasets.snli.load_data('train', 'ranking')
    x, y = train_data.unpack()
    assert len(x['text_left']) == num_samples
    assert len(x['text_right']) == num_samples
    assert y.shape == (num_samples, 1)


@pytest.mark.slow
def test_load_quora_qp():
    train_data = mz.datasets.quora_qp.load_data(task='classification')
    assert len(train_data) == 363177

    dev_data, tag = mz.datasets.quora_qp.load_data(
        'dev',
        task='classification',
        return_classes=True)
    assert tag == [False, True]
    assert len(dev_data) == 40371
    x, y = dev_data.unpack()
    assert len(x['text_left']) == 40371
    assert len(x['text_right']) == 40371
    assert y.shape == (40371, 2)

    test_data = mz.datasets.quora_qp.load_data('test')
    assert len(test_data) == 390965

    dev_data = mz.datasets.quora_qp.load_data('dev', 'ranking')
    x, y = dev_data.unpack()
    assert y.shape == (40371, 1)




"""One hot vectors."""
import numpy as np


def one_hot(indices: int, num_classes: int) -> np.ndarray:
    """:return: A one-hot encoded vector."""
    vec = np.zeros((num_classes,), dtype=np.int64)
    vec[indices] = 1
    return vec



"""The rank cross entropy loss."""
import numpy as np

from keras import layers, backend as K


class RankCrossEntropyLoss(object):
    """
    Rank cross entropy loss.

    Examples:
        >>> from keras import backend as K
        >>> softmax = lambda x: np.exp(x)/np.sum(np.exp(x), axis=0)
        >>> x_pred = K.variable(np.array([[1.0], [1.2], [0.8]]))
        >>> x_true = K.variable(np.array([[1], [0], [0]]))
        >>> expect = -np.log(softmax(np.array([[1.0], [1.2], [0.8]])))
        >>> loss = K.eval(RankCrossEntropyLoss(num_neg=2)(x_true, x_pred))
        >>> np.isclose(loss, expect[0]).all()
        True

    """

    def __init__(self, num_neg: int = 1):
        """
        :class:`RankCrossEntropyLoss` constructor.

        :param num_neg: number of negative instances in cross entropy loss.
        """
        self._num_neg = num_neg

    def __call__(self, y_true: np.array, y_pred: np.array) -> np.array:
        """
        Calculate rank cross entropy loss.

        :param y_true: Label.
        :param y_pred: Predicted result.
        :return: Crossentropy loss computed by user-defined negative number.
        """
        logits = layers.Lambda(lambda a: a[::(self._num_neg + 1), :])(y_pred)
        labels = layers.Lambda(lambda a: a[::(self._num_neg + 1), :])(y_true)
        logits, labels = [logits], [labels]
        for neg_idx in range(self._num_neg):
            neg_logits = layers.Lambda(
                lambda a: a[neg_idx + 1::(self._num_neg + 1), :])(y_pred)
            neg_labels = layers.Lambda(
                lambda a: a[neg_idx + 1::(self._num_neg + 1), :])(y_true)
            logits.append(neg_logits)
            labels.append(neg_labels)
        logits = K.concatenate(logits, axis=-1)
        labels = K.concatenate(labels, axis=-1)
        return -K.mean(K.sum(labels * K.log(K.softmax(logits)), axis=-1))

    @property
    def num_neg(self):
        """`num_neg` getter."""
        return self._num_neg



import typing
from pathlib import Path
import uuid

import matchzoo as mz
from matchzoo.engine.base_model import BaseModel
from .callback import Callback


class SaveModel(Callback):
    """
    Save trained model.

    For each trained model, a UUID will be generated as the `model_id`, the
    model will be saved under the `dir_path/model_id`. A `model_id` key will
    also be inserted into the result, which will visible in the return value of
    the `tune` method.

    :param dir_path: Path to save the models to. (default:
        `matchzoo.USER_TUNED_MODELS_DIR`)

    """

    def __init__(self, dir_path: typing.Union[str, Path]):
        """Init."""
        self._dir_path = dir_path or mz.USER_TUNED_MODELS_DIR

    def on_run_end(self, tuner, model: BaseModel, result: dict):
        """Save model on run end."""
        model_id = str(uuid.uuid4())
        model.save(self._dir_path.joinpath(model_id))
        result['model_id'] = model_id



import copy

import pytest
import keras

import matchzoo as mz


@pytest.fixture(scope='module')
def data_gen():
    return mz.DataGenerator(mz.datasets.toy.load_data())


@pytest.mark.parametrize('attr', [
    'callbacks',
    'num_neg',
    'num_dup',
    'mode',
    'batch_size',
    'shuffle',

])
def test_data_generator_getters_setters(data_gen, attr):
    assert hasattr(data_gen, attr)
    val = getattr(data_gen, attr)
    setattr(data_gen, attr, val)
    assert getattr(data_gen, attr) == val


def test_resample():
    model = mz.models.Naive()
    prpr = model.get_default_preprocessor()
    data_raw = mz.datasets.toy.load_data()
    data = prpr.fit_transform(data_raw)
    model.params.update(prpr.context)
    model.params['task'] = mz.tasks.Ranking()
    model.build()
    model.compile()

    data_gen = mz.DataGenerator(
        data_pack=data,
        mode='pair',
        resample=True,
        batch_size=4
    )

    class CheckResample(keras.callbacks.Callback):
        def __init__(self, data_gen):
            super().__init__()
            self._data_gen = data_gen
            self._orig_indices = None
            self._flags = []

        def on_epoch_end(self, epoch, logs=None):
            curr_indices = self._data_gen.batch_indices
            if not self._orig_indices:
                self._orig_indices = copy.deepcopy(curr_indices)
            else:
                self._flags.append(self._orig_indices != curr_indices)
                self._orig_indices = curr_indices

    check_resample = CheckResample(data_gen)
    model.fit_generator(data_gen, epochs=5, callbacks=[check_resample])
    assert check_resample._flags
    assert all(check_resample._flags)



"""The rank hinge loss."""
import numpy as np

from keras import layers, backend as K


class RankHingeLoss(object):
    """
    Rank hinge loss.

    Examples:
        >>> from keras import backend as K
        >>> x_pred = K.variable(np.array([[1.0], [1.2], [0.8], [1.4]]))
        >>> x_true = K.variable(np.array([[1], [0], [1], [0]]))
        >>> expect = ((1.0 + 1.2 - 1.0) + (1.0 + 1.4 - 0.8)) / 2
        >>> expect
        1.4
        >>> loss = K.eval(RankHingeLoss(num_neg=1, margin=1.0)(x_true, x_pred))
        >>> np.isclose(loss, expect)
        True

    """

    def __init__(self, num_neg: int = 1, margin: float = 1.0):
        """
        :class:`RankHingeLoss` constructor.

        :param num_neg: number of negative instances in hinge loss.
        :param margin: the margin between positive and negative scores.
        """
        self._num_neg = num_neg
        self._margin = margin

    def __call__(self, y_true: np.array, y_pred: np.array) -> np.array:
        """
        Calculate rank hinge loss.

        :param y_true: Label.
        :param y_pred: Predicted result.
        :return: Hinge loss computed by user-defined margin.
        """
        y_pos = layers.Lambda(lambda a: a[::(self._num_neg + 1), :],
                              output_shape=(1,))(y_pred)
        y_neg = []
        for neg_idx in range(self._num_neg):
            y_neg.append(
                layers.Lambda(
                    lambda a: a[(neg_idx + 1)::(self._num_neg + 1), :],
                    output_shape=(1,))(y_pred))
        y_neg = K.mean(K.concatenate(y_neg, axis=-1), axis=-1, keepdims=True)
        loss = K.maximum(0., self._margin + y_neg - y_pos)
        return K.mean(loss)

    @property
    def num_neg(self):
        """`num_neg` getter."""
        return self._num_neg

    @property
    def margin(self):
        """`margin` getter."""
        return self._margin



"""An implementation of ArcII Model."""
import typing

import keras

import matchzoo
from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class ArcII(BaseModel):
    """
    ArcII Model.

    Examples:
        >>> model = ArcII()
        >>> model.params['embedding_output_dim'] = 300
        >>> model.params['num_blocks'] = 2
        >>> model.params['kernel_1d_count'] = 32
        >>> model.params['kernel_1d_size'] = 3
        >>> model.params['kernel_2d_count'] = [16, 32]
        >>> model.params['kernel_2d_size'] = [[3, 3], [3, 3]]
        >>> model.params['pool_2d_size'] = [[2, 2], [2, 2]]
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(with_embedding=True)
        params['optimizer'] = 'adam'
        opt_space = hyper_spaces.choice(['adam', 'rmsprop', 'adagrad'])
        params.get('optimizer').hyper_space = opt_space
        params.add(Param(name='num_blocks', value=1,
                         desc="Number of 2D convolution blocks."))
        params.add(Param(name='kernel_1d_count', value=32,
                         desc="Kernel count of 1D convolution layer."))
        params.add(Param(name='kernel_1d_size', value=3,
                         desc="Kernel size of 1D convolution layer."))
        params.add(Param(name='kernel_2d_count', value=[32],
                         desc="Kernel count of 2D convolution layer in"
                              "each block"))
        params.add(Param(name='kernel_2d_size', value=[[3, 3]],
                         desc="Kernel size of 2D convolution layer in"
                              " each block."))
        params.add(Param(name='activation', value='relu',
                         desc="Activation function."))
        params.add(Param(name='pool_2d_size', value=[[2, 2]],
                         desc="Size of pooling layer in each block."))
        params.add(Param(
            name='padding', value='same',
            hyper_space=hyper_spaces.choice(
                ['same', 'valid']),
            desc="The padding mode in the convolution layer. It should be one"
                 "of `same`, `valid`."
        ))
        params.add(Param(
            name='dropout_rate', value=0.0,
            hyper_space=hyper_spaces.quniform(low=0.0, high=0.8,
                                              q=0.01),
            desc="The dropout rate."
        ))
        return params

    def build(self):
        """
        Build model structure.

        ArcII has the desirable property of letting two sentences meet before
        their own high-level representations mature.
        """
        input_left, input_right = self._make_inputs()

        embedding = self._make_embedding_layer()
        embed_left = embedding(input_left)
        embed_right = embedding(input_right)

        # Phrase level representations
        conv_1d_left = keras.layers.Conv1D(
            self._params['kernel_1d_count'],
            self._params['kernel_1d_size'],
            padding=self._params['padding']
        )(embed_left)
        conv_1d_right = keras.layers.Conv1D(
            self._params['kernel_1d_count'],
            self._params['kernel_1d_size'],
            padding=self._params['padding']
        )(embed_right)

        # Interaction
        matching_layer = matchzoo.layers.MatchingLayer(matching_type='plus')
        embed_cross = matching_layer([conv_1d_left, conv_1d_right])

        for i in range(self._params['num_blocks']):
            embed_cross = self._conv_pool_block(
                embed_cross,
                self._params['kernel_2d_count'][i],
                self._params['kernel_2d_size'][i],
                self._params['padding'],
                self._params['activation'],
                self._params['pool_2d_size'][i]
            )

        embed_flat = keras.layers.Flatten()(embed_cross)
        x = keras.layers.Dropout(rate=self._params['dropout_rate'])(embed_flat)

        inputs = [input_left, input_right]
        x_out = self._make_output_layer()(x)
        self._backend = keras.Model(inputs=inputs, outputs=x_out)

    @classmethod
    def _conv_pool_block(
        cls, x,
        kernel_count: int,
        kernel_size: int,
        padding: str,
        activation: str,
        pool_size: int
    ) -> typing.Any:
        output = keras.layers.Conv2D(kernel_count,
                                     kernel_size,
                                     padding=padding,
                                     activation=activation)(x)
        output = keras.layers.MaxPooling2D(pool_size=pool_size)(output)
        return output



"""A simple densely connected baseline model."""

import keras.layers

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class DenseBaseline(BaseModel):
    """
    A simple densely connected baseline model.

    Examples:
        >>> model = DenseBaseline()
        >>> model.params['mlp_num_layers'] = 2
        >>> model.params['mlp_num_units'] = 300
        >>> model.params['mlp_num_fan_out'] = 128
        >>> model.params['mlp_activation_func'] = 'relu'
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()
        >>> model.compile()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(with_multi_layer_perceptron=True)
        params['mlp_num_units'] = 256
        params.get('mlp_num_units').hyper_space = \
            hyper_spaces.quniform(16, 512)
        params.get('mlp_num_layers').hyper_space = \
            hyper_spaces.quniform(1, 5)
        return params

    def build(self):
        """Model structure."""
        x_in = self._make_inputs()
        x = keras.layers.concatenate(x_in)
        x = self._make_multi_layer_perceptron_layer()(x)
        x_out = self._make_output_layer()(x)
        self._backend = keras.models.Model(inputs=x_in, outputs=x_out)



import numpy as np

from .unit import Unit


class MatchingHistogram(Unit):
    """
    MatchingHistogramUnit Class.

    :param bin_size: The number of bins of the matching histogram.
    :param embedding_matrix: The word embedding matrix applied to calculate
                             the matching histogram.
    :param normalize: Boolean, normalize the embedding or not.
    :param mode: The type of the historgram, it should be one of 'CH', 'NG',
                 or 'LCH'.

    Examples:
        >>> embedding_matrix = np.array([[1.0, -1.0], [1.0, 2.0], [1.0, 3.0]])
        >>> text_left = [0, 1]
        >>> text_right = [1, 2]
        >>> histogram = MatchingHistogram(3, embedding_matrix, True, 'CH')
        >>> histogram.transform([text_left, text_right])
        [[3.0, 1.0, 1.0], [1.0, 2.0, 2.0]]

    """

    def __init__(self, bin_size: int = 30, embedding_matrix=None,
                 normalize=True, mode: str = 'LCH'):
        """The constructor."""
        self._hist_bin_size = bin_size
        self._embedding_matrix = embedding_matrix
        if normalize:
            self._normalize_embedding()
        self._mode = mode

    def _normalize_embedding(self):
        """Normalize the embedding matrix."""
        l2_norm = np.sqrt(
            (self._embedding_matrix * self._embedding_matrix).sum(axis=1)
        )
        self._embedding_matrix = \
            self._embedding_matrix / l2_norm[:, np.newaxis]

    def transform(self, input_: list) -> list:
        """Transform the input text."""
        text_left, text_right = input_
        matching_hist = np.ones((len(text_left), self._hist_bin_size),
                                dtype=np.float32)
        embed_left = self._embedding_matrix[text_left]
        embed_right = self._embedding_matrix[text_right]
        matching_matrix = embed_left.dot(np.transpose(embed_right))
        for (i, j), value in np.ndenumerate(matching_matrix):
            bin_index = int((value + 1.) / 2. * (self._hist_bin_size - 1.))
            matching_hist[i][bin_index] += 1.0
        if self._mode == 'NH':
            matching_sum = matching_hist.sum(axis=1)
            matching_hist = matching_hist / matching_sum[:, np.newaxis]
        elif self._mode == 'LCH':
            matching_hist = np.log(matching_hist)
        return matching_hist.tolist()



from pathlib import Path

USER_DIR = Path.expanduser(Path('~')).joinpath('.matchzoo')
if not USER_DIR.exists():
    USER_DIR.mkdir()
USER_DATA_DIR = USER_DIR.joinpath('datasets')
if not USER_DATA_DIR.exists():
    USER_DATA_DIR.mkdir()
USER_TUNED_MODELS_DIR = USER_DIR.joinpath('tuned_models')

from .version import __version__

from .data_pack import DataPack
from .data_pack import pack
from .data_pack import load_data_pack

from . import metrics
from . import tasks

from . import preprocessors
from . import data_generator
from .data_generator import DataGenerator
from .data_generator import DataGeneratorBuilder

from .preprocessors.chain_transform import chain_transform

from . import metrics
from . import losses
from . import engine
from . import models
from . import embedding
from . import datasets
from . import layers
from . import auto
from . import contrib

from .engine import hyper_spaces
from .engine.base_model import load_model
from .engine.base_preprocessor import load_preprocessor
from .engine import callbacks
from .engine.param import Param
from .engine.param_table import ParamTable

from .embedding.embedding import Embedding

from .utils import one_hot
from .preprocessors.build_unit_from_data_pack import build_unit_from_data_pack
from .preprocessors.build_vocab_unit import build_vocab_unit

# deprecated, should be removed in v2.2
from .contrib.legacy_data_generator import DPoolDataGenerator
from .contrib.legacy_data_generator import DPoolPairDataGenerator
from .contrib.legacy_data_generator import HistogramDataGenerator
from .contrib.legacy_data_generator import HistogramPairDataGenerator
from .contrib.legacy_data_generator import DynamicDataGenerator
from .contrib.legacy_data_generator import PairDataGenerator



"""Hyper parameter search spaces wrapping `hyperopt`."""
import typing
import numbers

import hyperopt
import hyperopt.pyll.base


class HyperoptProxy(object):
    """
    Hyperopt proxy class.

    See `hyperopt`'s documentation for more details:
    https://github.com/hyperopt/hyperopt/wiki/FMin

    Reason of these wrappers:

        A hyper space in `hyperopt` requires a `label` to instantiate. This
        `label` is used later as a reference to original hyper space that is
        sampled. In `matchzoo`, hyper spaces are used in
        :class:`matchzoo.engine.Param`. Only if a hyper space's label
        matches its parent :class:`matchzoo.engine.Param`'s name, `matchzoo`
        can correctly back-refrenced the parameter got sampled. This can be
        done by asking the user always use the same name for a parameter and
        its hyper space, but typos can occur. As a result, these wrappers
        are created to hide hyper spaces' `label`, and always correctly
        bind them with its parameter's name.

    Examples::
        >>> import matchzoo as mz
        >>> from hyperopt.pyll.stochastic import sample

    Basic Usage:
        >>> model = mz.models.DenseBaseline()
        >>> sample(model.params.hyper_space)  # doctest: +SKIP
         {'mlp_num_layers': 1.0, 'mlp_num_units': 274.0}

    Arithmetic Operations:
        >>> new_space = 2 ** mz.hyper_spaces.quniform(2, 6)
        >>> model.params.get('mlp_num_layers').hyper_space = new_space
        >>> sample(model.params.hyper_space)  # doctest: +SKIP
        {'mlp_num_layers': 8.0, 'mlp_num_units': 292.0}

    """

    def __init__(
        self,
        hyperopt_func: typing.Callable[..., hyperopt.pyll.Apply],
        **kwargs
    ):
        """
        :class:`HyperoptProxy` constructor.

        :param hyperopt_func: Target `hyperopt.hp` function to proxy.
        :param kwargs: Keyword arguments of the proxy function, must pass all
            parameters in `hyperopt_func`.
        """
        self._func = hyperopt_func
        self._kwargs = kwargs

    def convert(self, name: str) -> hyperopt.pyll.Apply:
        """
        Attach `name` as `hyperopt.hp`'s `label`.

        :param name:
        :return: a `hyperopt` ready search space
        """
        return self._func(name, **self._kwargs)

    def __add__(self, other):
        """__add__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x + y)

    def __radd__(self, other):
        """__radd__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x + y)

    def __sub__(self, other):
        """__sub__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x - y)

    def __rsub__(self, other):
        """__rsub__."""
        return _wrap_as_composite_func(self, other, lambda x, y: y - x)

    def __mul__(self, other):
        """__mul__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x * y)

    def __rmul__(self, other):
        """__rmul__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x * y)

    def __truediv__(self, other):
        """__truediv__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x / y)

    def __rtruediv__(self, other):
        """__rtruediv__."""
        return _wrap_as_composite_func(self, other, lambda x, y: y / x)

    def __floordiv__(self, other):
        """__floordiv__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x // y)

    def __rfloordiv__(self, other):
        """__rfloordiv__."""
        return _wrap_as_composite_func(self, other, lambda x, y: y // x)

    def __pow__(self, other):
        """__pow__."""
        return _wrap_as_composite_func(self, other, lambda x, y: x ** y)

    def __rpow__(self, other):
        """__rpow__."""
        return _wrap_as_composite_func(self, other, lambda x, y: y ** x)

    def __neg__(self):
        """__neg__."""
        return _wrap_as_composite_func(self, None, lambda x, _: -x)


def _wrap_as_composite_func(self, other, func):
    def _wrapper(name, **kwargs):
        return func(self._func(name, **kwargs), other)

    return HyperoptProxy(_wrapper, **self._kwargs)


class choice(HyperoptProxy):
    """:func:`hyperopt.hp.choice` proxy."""

    def __init__(self, options: list):
        """
        :func:`hyperopt.hp.choice` proxy.

        :param options: options to search from
        """
        super().__init__(hyperopt_func=hyperopt.hp.choice, options=options)
        self._options = options

    def __str__(self):
        """:return: `str` representation of the hyper space."""
        return f'choice in {self._options}'


class quniform(HyperoptProxy):
    """:func:`hyperopt.hp.quniform` proxy."""

    def __init__(
        self,
        low: numbers.Number,
        high: numbers.Number,
        q: numbers.Number = 1
    ):
        """
        :func:`hyperopt.hp.quniform` proxy.

        If using with integer values, then `high` is exclusive.

        :param low: lower bound of the space
        :param high: upper bound of the space
        :param q: similar to the `step` in the python built-in `range`
        """
        super().__init__(hyperopt_func=hyperopt.hp.quniform,
                         low=low,
                         high=high, q=q)
        self._low = low
        self._high = high
        self._q = q

    def __str__(self):
        """:return: `str` representation of the hyper space."""
        return f'quantitative uniform distribution in  ' \
               f'[{self._low}, {self._high}), with a step size of {self._q}'


class uniform(HyperoptProxy):
    """:func:`hyperopt.hp.uniform` proxy."""

    def __init__(
        self,
        low: numbers.Number,
        high: numbers.Number
    ):
        """
        :func:`hyperopt.hp.uniform` proxy.

        :param low: lower bound of the space
        :param high: upper bound of the space
        """
        super().__init__(hyperopt_func=hyperopt.hp.uniform, low=low, high=high)
        self._low = low
        self._high = high

    def __str__(self):
        """:return: `str` representation of the hyper space."""
        return f'uniform distribution in  [{self._low}, {self._high})'


def sample(space):
    """
    Take a sample in the hyper space.

    This method is stateless, so the distribution of the samples is different
    from that of `tune` call. This function just gives a general idea of what
    a sample from the `space` looks like.

    Example:
        >>> import matchzoo as mz
        >>> space = mz.models.Naive.get_default_params().hyper_space
        >>> mz.hyper_spaces.sample(space)  # doctest: +ELLIPSIS
        {'optimizer': ...}

    """
    return hyperopt.pyll.stochastic.sample(space)



import io
import os

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

# Avoids IDE errors, but actual version is read from version.py
__version__ = None
exec(open('matchzoo/version.py').read())

short_description = 'Facilitating the design, comparison and sharing of deep text matching models.'

# Get the long description from the README file
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    'keras >= 2.2.4',
    'nltk >= 3.2.3',
    'numpy >= 1.14',
    'tqdm >= 4.19.4',
    'dill >= 0.2.7.1',
    'pandas >= 0.23.1',
    'networkx >= 2.1',
    'h5py >= 2.8.0',
    'hyperopt >= 0.1.1'
]

extras_requires = {
    'tests': [
        'coverage >= 4.3.4',
        'codecov >= 2.0.15',
        'pytest >= 3.0.3',
        'pytest-cov >= 2.4.0',
        'flake8 >= 3.6.0',
        'flake8_docstrings >= 1.0.2'],
}


setup(
    name="MatchZoo",
    version=__version__,
    author="Yixing Fan, Bo Wang, Zeyi Wang, Liang Pang, Liu Yang, Qinghua Wang, etc.",
    author_email="fanyixing@ict.ac.cn",
    description=(short_description),
    license="Apache 2.0",
    keywords="text matching models",
    url="https://github.com/NTMC-Community/MatchZoo",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 3 - Alpha",
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        "License :: OSI Approved :: Apache Software License",
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=install_requires,
    extras_require=extras_requires
)



"""Wrapper function organizes a number of transform functions."""
import typing
import functools

from .units.unit import Unit


def chain_transform(units: typing.List[Unit]) -> typing.Callable:
    """
    Compose unit transformations into a single function.

    :param units: List of :class:`matchzoo.StatelessUnit`.
    """

    @functools.wraps(chain_transform)
    def wrapper(arg):
        """Wrapper function of transformations composition."""
        for unit in units:
            arg = unit.transform(arg)
        return arg

    unit_names = ' => '.join(unit.__class__.__name__ for unit in units)
    wrapper.__name__ += ' of ' + unit_names
    return wrapper



"""Normalized discounted cumulative gain metric for ranking."""
import numpy as np

from matchzoo.engine.base_metric import BaseMetric, sort_and_couple
from .discounted_cumulative_gain import DiscountedCumulativeGain


class NormalizedDiscountedCumulativeGain(BaseMetric):
    """Normalized discounted cumulative gain metric."""

    ALIAS = ['normalized_discounted_cumulative_gain', 'ndcg']

    def __init__(self, k: int = 1, threshold: float = 0.):
        """
        :class:`NormalizedDiscountedCumulativeGain` constructor.

        :param k: Number of results to consider
        :param threshold: the label threshold of relevance degree.
        """
        self._k = k
        self._threshold = threshold

    def __repr__(self) -> str:
        """:return: Formated string representation of the metric."""
        return f"{self.ALIAS[0]}@{self._k}({self._threshold})"

    def __call__(self, y_true: np.array, y_pred: np.array) -> float:
        """
        Calculate normalized discounted cumulative gain (ndcg).

        Relevance is positive real values or binary values.

        Example:
            >>> y_true = [0, 1, 2, 0]
            >>> y_pred = [0.4, 0.2, 0.5, 0.7]
            >>> ndcg = NormalizedDiscountedCumulativeGain
            >>> ndcg(k=1)(y_true, y_pred)
            0.0
            >>> round(ndcg(k=2)(y_true, y_pred), 2)
            0.52
            >>> round(ndcg(k=3)(y_true, y_pred), 2)
            0.52
            >>> type(ndcg()(y_true, y_pred))
            <class 'float'>

        :param y_true: The ground true label of each document.
        :param y_pred: The predicted scores of each document.

        :return: Normalized discounted cumulative gain.
        """
        dcg_metric = DiscountedCumulativeGain(k=self._k,
                                              threshold=self._threshold)
        idcg_val = dcg_metric(y_true, y_true)
        dcg_val = dcg_metric(y_true, y_pred)
        return dcg_val / idcg_val if idcg_val != 0 else 0



import numpy as np
import pytest
from keras import backend as K

from matchzoo import layers


def test_matching_layers():
    s1_value = np.array([[[1, 2], [2, 3], [3, 4]],
                         [[0.1, 0.2], [0.2, 0.3], [0.3, 0.4]]
                        ])
    s2_value = np.array([[[1, 2], [2, 3]],
                         [[0.1, 0.2], [0.2, 0.3]]
                        ])
    s3_value = np.array([[[1, 2], [2, 3]],
                         [[0.1, 0.2], [0.2, 0.3]],
                         [[0.1, 0.2], [0.2, 0.3]]
                        ])
    s1_tensor = K.variable(s1_value)
    s2_tensor = K.variable(s2_value)
    s3_tensor = K.variable(s3_value)
    for matching_type in ['dot', 'mul', 'plus', 'minus', 'concat']:
        model = layers.MatchingLayer(matching_type=matching_type)([s1_tensor, s2_tensor])
        ret = K.eval(model)
    with pytest.raises(ValueError):
        layers.MatchingLayer(matching_type='error')
    with pytest.raises(ValueError):
        layers.MatchingLayer()([s1_tensor, s3_tensor])



from matchzoo.engine.base_model import BaseModel
from .callback import Callback


class LoadEmbeddingMatrix(Callback):
    """
    Load a pre-trained embedding after the model is built.

    :param embedding_matrix: Embedding matrix to load.

    """

    def __init__(self, embedding_matrix):
        """Init."""
        self._embedding_matrix = embedding_matrix

    def on_build_end(self, tuner, model: BaseModel):
        """`on_build_end`."""
        model.load_embedding_matrix(self._embedding_matrix)



from . import toy
from . import wiki_qa
from . import embeddings
from . import snli
from . import quora_qp
from pathlib import Path


def list_available():
    return [p.name for p in Path(__file__).parent.iterdir()
            if p.is_dir() and not p.name.startswith('_')]



"""Naive Preprocessor."""

from tqdm import tqdm

from matchzoo.engine.base_preprocessor import BasePreprocessor
from matchzoo import DataPack
from .chain_transform import chain_transform
from .build_vocab_unit import build_vocab_unit
from . import units

tqdm.pandas()


class NaivePreprocessor(BasePreprocessor):
    """
    Naive preprocessor.

    Example:
        >>> import matchzoo as mz
        >>> train_data = mz.datasets.toy.load_data()
        >>> test_data = mz.datasets.toy.load_data(stage='test')
        >>> preprocessor = mz.preprocessors.NaivePreprocessor()
        >>> train_data_processed = preprocessor.fit_transform(train_data,
        ...                                                   verbose=0)
        >>> type(train_data_processed)
        <class 'matchzoo.data_pack.data_pack.DataPack'>
        >>> test_data_transformed = preprocessor.transform(test_data,
        ...                                                verbose=0)
        >>> type(test_data_transformed)
        <class 'matchzoo.data_pack.data_pack.DataPack'>

    """

    def fit(self, data_pack: DataPack, verbose: int = 1):
        """
        Fit pre-processing context for transformation.

        :param data_pack: data_pack to be preprocessed.
        :param verbose: Verbosity.
        :return: class:`NaivePreprocessor` instance.
        """
        func = chain_transform(self._default_units())
        data_pack = data_pack.apply_on_text(func, verbose=verbose)
        vocab_unit = build_vocab_unit(data_pack, verbose=verbose)
        self._context['vocab_unit'] = vocab_unit
        return self

    def transform(self, data_pack: DataPack, verbose: int = 1) -> DataPack:
        """
        Apply transformation on data, create `tri-letter` representation.

        :param data_pack: Inputs to be preprocessed.
        :param verbose: Verbosity.

        :return: Transformed data as :class:`DataPack` object.
        """
        units_ = self._default_units()
        units_.append(self._context['vocab_unit'])
        units_.append(units.FixedLength(text_length=30, pad_mode='post'))
        func = chain_transform(units_)
        return data_pack.apply_on_text(func, verbose=verbose)



"""An implementation of ArcI Model."""
import typing

import keras

from matchzoo.engine.base_model import BaseModel
from matchzoo.engine.param import Param
from matchzoo.engine.param_table import ParamTable
from matchzoo.engine import hyper_spaces


class ArcI(BaseModel):
    """
    ArcI Model.

    Examples:
        >>> model = ArcI()
        >>> model.params['num_blocks'] = 1
        >>> model.params['left_filters'] = [32]
        >>> model.params['right_filters'] = [32]
        >>> model.params['left_kernel_sizes'] = [3]
        >>> model.params['right_kernel_sizes'] = [3]
        >>> model.params['left_pool_sizes'] = [2]
        >>> model.params['right_pool_sizes'] = [4]
        >>> model.params['conv_activation_func'] = 'relu'
        >>> model.params['mlp_num_layers'] = 1
        >>> model.params['mlp_num_units'] = 64
        >>> model.params['mlp_num_fan_out'] = 32
        >>> model.params['mlp_activation_func'] = 'relu'
        >>> model.params['dropout_rate'] = 0.5
        >>> model.guess_and_fill_missing_params(verbose=0)
        >>> model.build()

    """

    @classmethod
    def get_default_params(cls) -> ParamTable:
        """:return: model default parameters."""
        params = super().get_default_params(
            with_embedding=True,
            with_multi_layer_perceptron=True
        )
        params['optimizer'] = 'adam'
        params.add(Param(name='num_blocks', value=1,
                         desc="Number of convolution blocks."))
        params.add(Param(name='left_filters', value=[32],
                         desc="The filter size of each convolution "
                              "blocks for the left input."))
        params.add(Param(name='left_kernel_sizes', value=[3],
                         desc="The kernel size of each convolution "
                              "blocks for the left input."))
        params.add(Param(name='right_filters', value=[32],
                         desc="The filter size of each convolution "
                              "blocks for the right input."))
        params.add(Param(name='right_kernel_sizes', value=[3],
                         desc="The kernel size of each convolution "
                              "blocks for the right input."))
        params.add(Param(name='conv_activation_func', value='relu',
                         desc="The activation function in the "
                              "convolution layer."))
        params.add(Param(name='left_pool_sizes', value=[2],
                         desc="The pooling size of each convolution "
                              "blocks for the left input."))
        params.add(Param(name='right_pool_sizes', value=[2],
                         desc="The pooling size of each convolution "
                              "blocks for the right input."))
        params.add(Param(
            name='padding',
            value='same',
            hyper_space=hyper_spaces.choice(
                ['same', 'valid', 'causal']),
            desc="The padding mode in the convolution layer. It should be one"
                 "of `same`, `valid`, and `causal`."
        ))
        params.add(Param(
            'dropout_rate', 0.0,
            hyper_space=hyper_spaces.quniform(
                low=0.0, high=0.8, q=0.01),
            desc="The dropout rate."
        ))
        return params

    def build(self):
        """
        Build model structure.

        ArcI use Siamese arthitecture.
        """
        input_left, input_right = self._make_inputs()

        embedding = self._make_embedding_layer()
        embed_left = embedding(input_left)
        embed_right = embedding(input_right)

        for i in range(self._params['num_blocks']):
            embed_left = self._conv_pool_block(
                embed_left,
                self._params['left_filters'][i],
                self._params['left_kernel_sizes'][i],
                self._params['padding'],
                self._params['conv_activation_func'],
                self._params['left_pool_sizes'][i]
            )
            embed_right = self._conv_pool_block(
                embed_right,
                self._params['right_filters'][i],
                self._params['right_kernel_sizes'][i],
                self._params['padding'],
                self._params['conv_activation_func'],
                self._params['right_pool_sizes'][i]
            )

        rep_left = keras.layers.Flatten()(embed_left)
        rep_right = keras.layers.Flatten()(embed_right)
        concat = keras.layers.Concatenate(axis=1)([rep_left, rep_right])
        dropout = keras.layers.Dropout(
            rate=self._params['dropout_rate'])(concat)
        mlp = self._make_multi_layer_perceptron_layer()(dropout)

        inputs = [input_left, input_right]
        x_out = self._make_output_layer()(mlp)
        self._backend = keras.Model(inputs=inputs, outputs=x_out)

    def _conv_pool_block(
        self,
        input_: typing.Any,
        filters: int,
        kernel_size: int,
        padding: str,
        conv_activation_func: str,
        pool_size: int
    ) -> typing.Any:
        output = keras.layers.Conv1D(
            filters,
            kernel_size,
            padding=padding,
            activation=conv_activation_func
        )(input_)
        output = keras.layers.MaxPooling1D(pool_size=pool_size)(output)
        return output



import typing

import matchzoo
from matchzoo.engine.base_metric import BaseMetric
from matchzoo.engine import base_task


def parse_metric(
    metric: typing.Union[str, typing.Type[BaseMetric], BaseMetric],
    task: 'base_task.BaseTask' = None
) -> typing.Union['BaseMetric', str]:
    """
    Parse input metric in any form into a :class:`BaseMetric` instance.

    :param metric: Input metric in any form.
    :param task: Task type for determining specific metric.
    :return: A :class:`BaseMetric` instance

    Examples::
        >>> from matchzoo import metrics
        >>> from matchzoo.engine.parse_metric import parse_metric

    Use `str` as keras native metrics:
        >>> parse_metric('mse')
        'mse'

    Use `str` as MatchZoo metrics:
        >>> mz_metric = parse_metric('map')
        >>> type(mz_metric)
        <class 'matchzoo.metrics.mean_average_precision.MeanAveragePrecision'>

    Use :class:`matchzoo.engine.BaseMetric` subclasses as MatchZoo metrics:
        >>> type(parse_metric(metrics.AveragePrecision))
        <class 'matchzoo.metrics.average_precision.AveragePrecision'>

    Use :class:`matchzoo.engine.BaseMetric` instances as MatchZoo metrics:
        >>> type(parse_metric(metrics.AveragePrecision()))
        <class 'matchzoo.metrics.average_precision.AveragePrecision'>

    """
    if task is None:
        task = matchzoo.tasks.Ranking()

    if isinstance(metric, str):
        metric = metric.lower()  # ignore case

        # matchzoo metrics in str form
        for subclass in BaseMetric.__subclasses__():
            if metric == subclass.ALIAS or metric in subclass.ALIAS:
                return subclass()

        # keras native metrics
        return _remap_keras_metric(metric, task)
    elif isinstance(metric, BaseMetric):
        return metric
    elif issubclass(metric, BaseMetric):
        return metric()
    else:
        raise ValueError(metric)


def _remap_keras_metric(metric: str, task) -> str:
    # we do not support sparse label in classification.
    lookup = {
        matchzoo.tasks.Ranking: {
            'acc': 'binary_accuracy',
            'accuracy': 'binary_accuracy',
            'crossentropy': 'binary_crossentropy',
            'ce': 'binary_crossentropy',
        },
        matchzoo.tasks.Classification: {
            'acc': 'categorical_accuracy',
            'accuracy': 'categorical_accuracy',
            'crossentropy': 'categorical_crossentropy',
            'ce': 'categorical_crossentropy',
        }
    }
    return lookup[type(task)].get(metric, metric)



"""Convert list of input into class:`DataPack` expected format."""

import typing

import pandas as pd
import numpy as np

import matchzoo


def pack(df: pd.DataFrame) -> 'matchzoo.DataPack':
    """
    Pack a :class:`DataPack` using `df`.

    The `df` must have `text_left` and `text_right` columns. Optionally,
    the `df` can have `id_left`, `id_right` to index `text_left` and
    `text_right` respectively. `id_left`, `id_right` will be automatically
    generated if not specified.

    :param df: Input :class:`pandas.DataFrame` to use.

    Examples::
        >>> import matchzoo as mz
        >>> import pandas as pd
        >>> df = pd.DataFrame(data={'text_left': list('AABC'),
        ...                         'text_right': list('abbc'),
        ...                         'label': [0, 1, 1, 0]})
        >>> mz.pack(df).frame()
          id_left text_left id_right text_right  label
        0     L-0         A      R-0          a      0
        1     L-0         A      R-1          b      1
        2     L-1         B      R-1          b      1
        3     L-2         C      R-2          c      0

    """
    if 'text_left' not in df or 'text_right' not in df:
        raise ValueError(
            'Input data frame must have `text_left` and `text_right`.')

    # Gather IDs
    if 'id_left' not in df:
        id_left = _gen_ids(df, 'text_left', 'L-')
    else:
        id_left = df['id_left']
    if 'id_right' not in df:
        id_right = _gen_ids(df, 'text_right', 'R-')
    else:
        id_right = df['id_right']

    # Build Relation
    relation = pd.DataFrame(data={'id_left': id_left, 'id_right': id_right})
    for col in df:
        if col not in ['id_left', 'id_right', 'text_left', 'text_right']:
            relation[col] = df[col]

    # Build Left and Right
    left = _merge(df, id_left, 'text_left', 'id_left')
    right = _merge(df, id_right, 'text_right', 'id_right')
    return matchzoo.DataPack(relation, left, right)


def _merge(data: pd.DataFrame, ids: typing.Union[list, np.array],
           text_label: str, id_label: str):
    left = pd.DataFrame(data={
        text_label: data[text_label], id_label: ids
    })
    left.drop_duplicates(id_label, inplace=True)
    left.set_index(id_label, inplace=True)
    return left


def _gen_ids(data: pd.DataFrame, col: str, prefix: str):
    lookup = {}
    for text in data[col].unique():
        lookup[text] = prefix + str(len(lookup))
    return data[col].map(lookup)
